from logging import getLogger
from typing import Any, Dict, List, Optional

from watchmen_storage import SnowflakeGenerator
from watchmen_utilities import ArrayHelper, is_blank

MY_AID_ID = 'aid_me'
AID_ROOT = 'aid_root'

logger = getLogger(__name__)


class Ancestor:
	def __init__(self, name: str, aid_id: int):
		self.name = name
		self.aidId = aid_id


# noinspection GrazieInspection
def apply_ancestor_aid_id(
		data: Dict[str, Any], my_hierarchy_number: int,
		ancestor: Ancestor, used_ancestor_keys: Dict[str, bool],
) -> None:
	"""
	apply ancestor aid id to given data.\n
	if not applied(used keys is 0), apply as to root, use "aid_root" as name,\n
	use "aid_{ancestor.name}" as name when this name is not used,\n
	use "aid_{ancestor.name}_{distance_to_ancestor}" as name when name in step b is used.
	distance is the difference value of my hierarchy number and ancestor's hierarchy number

	for example:
	root.a.b.c.b.e, now data is e, will create the following aid properties:\n
	aid_root: to root,\n
	aid_a: to a,\n
	aid_b: to b,\n
	aid_c: to c,\n
	aid_b_1: to b which is closer to me(e).
	"""
	used_count = len(used_ancestor_keys)
	if len(used_ancestor_keys) == 0:
		# first one is always key of root
		name = AID_ROOT
	else:
		name = f'aid_{ancestor.name}'
		if name in used_ancestor_keys:
			name = f'aid_{ancestor.name}_{my_hierarchy_number - used_count}'
	data[name] = ancestor.aidId
	used_ancestor_keys[name] = True


def aid(
		data: Dict[str, Any], ancestors: Optional[List[Ancestor]], snowflake_generator: SnowflakeGenerator
) -> None:
	"""
	given data should be modified
	"""

	try:
		# create aid me
		aid_me = snowflake_generator.next_id()
		data[MY_AID_ID] = aid_me

		# create ancestor aid ids
		my_hierarchy_number = len(ancestors)
		used_ancestor_keys: Dict[str, bool] = {}
		ArrayHelper(ancestors).each(lambda x: apply_ancestor_aid_id(data, my_hierarchy_number, x, used_ancestor_keys))

		for key in data:
			value = data[key]
			my_ancestors = ArrayHelper(ancestors).copy().grab(Ancestor(name=key, aid_id=aid_me)).to_list()
			if isinstance(value, dict):
				aid(value, my_ancestors, snowflake_generator)
			elif isinstance(value, list):
				def aid_each(element: Any) -> None:
					if isinstance(element, dict):
						aid(element, my_ancestors, snowflake_generator)
					elif isinstance(element, list):
						ArrayHelper(element).each(aid_each)

				ArrayHelper(value).each(aid_each)
	except Exception as e:
		hierarchy = ArrayHelper(ancestors).map(lambda x: x.name).join('/')
		if is_blank(hierarchy):
			hierarchy = 'root'
		logger.error(f'Error occurred on hierarchy [{hierarchy}].')
		raise e



from re import findall
from typing import List, Tuple

from watchmen_data_kernel.common import DataKernelException
from watchmen_utilities import ArrayHelper, is_blank, is_not_blank


class MightAVariable:
	def __init__(self, text: str, variable: str):
		self.text = text
		self.variable = variable

	def has_text(self):
		return is_not_blank(self.text)

	def has_variable(self):
		return is_not_blank(self.variable)


def parse_variable(variable_str: str) -> Tuple[bool, List[MightAVariable]]:
	# parsed result is:
	# for empty string, a list contains one tuple: [('', '')]
	# for no variable string, a list contains 2 tuples: [(value, ''), ('', '')]
	# found, a list contains x tuples: [(first, first_variable), (second, second_variable), ..., ('', '')]
	parsed = findall('([^{]*({[^}]+})?)', variable_str)
	if parsed[0][0] == '':
		# no variable required
		return False, [MightAVariable(text=variable_str, variable='')]
	else:
		def beautify(a_tuple: Tuple[str, str]) -> MightAVariable:
			if a_tuple[1] == '':
				# no variable in it
				return MightAVariable(text=a_tuple[0], variable='')
			else:
				# remove variable from first, remove braces from second
				return MightAVariable(text=a_tuple[0][: (0 - len(a_tuple[1]))], variable=a_tuple[1][1:-1].strip())

		return True, ArrayHelper(parsed[:-1]).map(lambda x: beautify(x)).to_list()


def parse_function_in_variable(function_literal: str, function_name: str, parameters_count: int):
	if parameters_count < 1:
		raise DataKernelException('At least one parameter for function in variable.')

	one_param = ['(.+)']
	all_params = ','.join([param for param in one_param for _ in range(parameters_count)])
	parsed_params = findall(f'^({function_name})\\s*\\({all_params}\\)$', function_literal)
	if len(parsed_params) != 1:
		raise DataKernelException(f'Constant[{function_literal}] is not supported.')
	if len(parsed_params[0]) != parameters_count + 1:
		raise DataKernelException(f'Constant[{function_literal}] is not supported.')
	variable_names = parsed_params[0][1:]
	for variable_name in variable_names:
		if is_blank(variable_name):
			raise DataKernelException(f'Constant[{function_literal}] is not supported.')
	return variable_names


def parse_move_date_pattern(pattern: str) -> List[Tuple[str, str, str]]:
	"""
	elements of tuple are
	1. flag: YMDhms
	2. operator: +- or empty
	3. number
	"""
	pattern = pattern.strip()
	segments = findall(r'([YMDhms])\s*([+-]?)\s*(\d+)', pattern)
	parsed = ArrayHelper(segments).map(lambda x: f'{x[0]}{x[1]}{x[2]}').join('')
	original = ArrayHelper([*pattern]).filter(lambda x: is_not_blank(x)).join('')
	if parsed != original:
		raise DataKernelException(f'Incorrect date move pattern[{pattern}].')
	return segments



from typing import Any, Dict

from cacheout import Cache, CacheManager


# noinspection DuplicatedCode
class TopicByIdCache(Cache):
	pass


class TopicByTenantAndNameCache(Cache):
	pass


class TopicSchemaByIdCache(Cache):
	pass


class TopicEntityHelperByIdCache(Cache):
	pass


class PipelineByIdCache(Cache):
	pass


class PipelineByTopicIdCache(Cache):
	pass


class DataSourceByIdCache(Cache):
	pass


class DataStorageBuilderByIdCache(Cache):
	pass


class ExternalWriterByIdCache(Cache):
	pass


class TenantByIdCache(Cache):
	pass


class KeyStoreByTypeCache(Cache):
	pass


cache_set = CacheManager()
cache_set.setup({
	'TOPIC_BY_ID': {'cache_class': TopicByIdCache, 'maxsize': 512},
	'TOPIC_BY_TENANT_AND_NAME': {'cache_class': TopicByTenantAndNameCache, 'maxsize': 512},
	'TOPIC_SCHEMA_BY_ID': {'cache_class': TopicSchemaByIdCache, 'maxsize': 512},
	'TOPIC_ENTITY_HELPER_BY_ID': {'cache_class': TopicEntityHelperByIdCache, 'maxsize': 512},
	'PIPELINE_BY_ID': {'cache_class': PipelineByIdCache, 'maxsize': 1024},
	'PIPELINE_BY_TOPIC_ID': {'cache_class': PipelineByTopicIdCache, 'maxsize': 512},
	'DATA_SOURCE_BY_ID': {'cache_class': DataSourceByIdCache, 'maxsize': 64},
	'DATA_STORAGE_BUILDER_BY_ID': {'cache_class': DataStorageBuilderByIdCache, 'maxsize': 64},
	'EXTERNAL_WRITER_BY_ID': {'cache_class': ExternalWriterByIdCache, 'maxsize': 16},
	'TENANT_BY_ID': {'cache_class': TenantByIdCache, 'maxsize': 32},
	'KEY_STORE_BY_TYPE': {'cache_class': KeyStoreByTypeCache, 'maxsize': 32}
})


def configure_cache(name: str, options: Dict[str, Any]) -> None:
	cache_set.configure(name, **options)


def find_cache(name: str) -> Cache:
	return cache_set[name]


def get_topic_by_id_cache() -> TopicByIdCache:
	return cache_set['TOPIC_BY_ID']


def get_topic_by_tenant_and_name_cache() -> TopicByTenantAndNameCache:
	return cache_set['TOPIC_BY_TENANT_AND_NAME']


def get_topic_schema_by_id_cache() -> TopicSchemaByIdCache:
	return cache_set['TOPIC_SCHEMA_BY_ID']


def get_topic_entity_helper_by_id_cache() -> TopicEntityHelperByIdCache:
	return cache_set['TOPIC_ENTITY_HELPER_BY_ID']


def get_pipeline_by_id_cache() -> PipelineByIdCache:
	return cache_set['PIPELINE_BY_ID']


def get_pipeline_by_topic_id_cache() -> PipelineByTopicIdCache:
	return cache_set['PIPELINE_BY_TOPIC_ID']


def get_data_source_by_id_cache() -> DataSourceByIdCache:
	return cache_set['DATA_SOURCE_BY_ID']


def get_data_storage_builder_by_id_cache() -> DataStorageBuilderByIdCache:
	return cache_set['DATA_STORAGE_BUILDER_BY_ID']


def get_external_writer_by_id_cache() -> ExternalWriterByIdCache:
	return cache_set['EXTERNAL_WRITER_BY_ID']


def get_tenant_by_id_cache() -> TenantByIdCache:
	return cache_set['TENANT_BY_ID']


def get_key_store_by_type_cache() -> KeyStoreByTypeCache:
	return cache_set['KEY_STORE_BY_TYPE']



from typing import Any, Union

from watchmen_data_kernel.common import DataKernelException
from watchmen_model.admin import FactorEncryptMethod
from .encryptor import Encryptor


class LastMasker(Encryptor):
	def __init__(self, digits: int):
		self.digits = digits
		if digits == 3:
			self.method = FactorEncryptMethod.MASK_LAST_3
		elif digits == 6:
			self.method = FactorEncryptMethod.MASK_LAST_6
		else:
			raise DataKernelException(f'Only 3 or 6 digits last mask is supported, current is [{digits}].')

	def accept(self, method: Union[FactorEncryptMethod, str]) -> bool:
		return method == self.method

	def is_encrypted(self, value: Any) -> bool:
		"""
		always returns false
		"""
		return False

	def do_encrypt(self, value: Any) -> Any:
		value = str(value)
		length = len(value)
		if length <= self.digits:
			# mask all
			return '*' * length
		decimal_count = len([ch for ch in value if ch.isdecimal()])
		if decimal_count < self.digits:
			# there are not enough decimal characters,
			# mask last n characters, no matter what it is
			return f'{value[0:length - self.digits]}{"*" * self.digits}'
		else:
			# mask last n decimal characters
			reversed_value = value[::-1]
			replaced_value = []
			masked_count = 0
			for ch in reversed_value:
				if masked_count < 3 and ch.isdecimal():
					masked_count = masked_count + 1
					replaced_value.append('*')
				else:
					replaced_value.append(ch)
			return ''.join(replaced_value)[::-1]

	def do_decrypt(self, value: str) -> str:
		# center mask cannot be decrypted
		return value

	def get_key_type(self) -> str:
		return self.method.value



from typing import List, Optional

from watchmen_auth import PrincipalService
from watchmen_data_kernel.cache import CacheService
from watchmen_data_kernel.common import DataKernelException
from watchmen_meta.admin import PipelineService as PipelineStorageService
from watchmen_meta.common import ask_meta_storage, ask_snowflake_generator
from watchmen_model.admin import Pipeline
from watchmen_model.common import PipelineId, TopicId
from watchmen_utilities import ArrayHelper


class PipelineService:
	def __init__(self, principal_service: PrincipalService):
		self.principalService = principal_service

	def find_by_id(self, pipeline_id: PipelineId) -> Optional[Pipeline]:
		pipeline = CacheService.pipeline().get(pipeline_id)
		if pipeline is not None:
			if pipeline.tenantId != self.principalService.get_tenant_id():
				raise DataKernelException(
					f'Pipeline[id={pipeline_id}] not belongs to '
					f'current tenant[id={self.principalService.get_tenant_id()}].')
			return pipeline

		storage_service = PipelineStorageService(ask_meta_storage(), ask_snowflake_generator(), self.principalService)
		storage_service.begin_transaction()
		try:
			# noinspection PyTypeChecker
			pipeline: Pipeline = storage_service.find_by_id(pipeline_id)
			if pipeline is None:
				return None

			CacheService.pipeline().put(pipeline)
			return pipeline
		finally:
			storage_service.close_transaction()

	def find_by_topic_id(self, topic_id: TopicId) -> List[Pipeline]:
		pipeline_ids = CacheService.pipelines_by_topic().get(topic_id)
		if pipeline_ids is not None:
			pipelines = ArrayHelper(pipeline_ids) \
				.map(lambda x: self.find_by_id(x)) \
				.filter(lambda x: x is not None).to_list()
			if len(pipelines) != len(pipeline_ids):
				loaded = ArrayHelper(pipelines).map(lambda x: x.pipelineId).to_list()
				raise Exception(f'Except pipelines[{pipeline_ids}], but get[{loaded}] only.')
			return pipelines

		storage_service = PipelineStorageService(ask_meta_storage(), ask_snowflake_generator(), self.principalService)
		storage_service.begin_transaction()
		try:
			# noinspection PyTypeChecker
			pipelines: List[Pipeline] = storage_service.find_by_topic_id(
				topic_id, self.principalService.get_tenant_id())
			if len(pipelines) == 0:
				CacheService.pipelines_by_topic().declare_no_pipelines(topic_id)
				return pipelines

			return ArrayHelper(pipelines).each(lambda x: CacheService.pipeline().put(x)).to_list()
		finally:
			storage_service.close_transaction()



from __future__ import annotations

from datetime import date
from decimal import Decimal
from typing import Any, Callable, Dict, List, Optional, Tuple

from watchmen_auth import PrincipalService
from watchmen_data_kernel.common import ask_all_date_formats, DataKernelException
from watchmen_meta.common import ask_snowflake_generator
from watchmen_model.common import VariablePredefineFunctions
from watchmen_utilities import ArrayHelper, get_current_time_in_seconds, is_date, month_diff, \
	truncate_time, try_to_decimal, year_diff
from .variables import PipelineVariables


def get_value_from(
		name: str, names: List[str],
		get_first: Callable[[str], Any], is_list: Callable[[List[str]], bool]) -> Any:
	data = get_first(names[0])
	if data is None:
		if is_list(names[:1]):
			return []
		else:
			return None

	remains_count: int = len(names) - 1
	current_index: int = 1
	while current_index <= remains_count:
		current_name = names[current_index]
		if current_name == VariablePredefineFunctions.COUNT:
			if isinstance(data, list) or isinstance(data, dict):
				return len(data)
			else:
				raise DataKernelException(f'Cannot retrieve[key={name}, current={current_name}] from [{data}].')
		elif current_name == VariablePredefineFunctions.LENGTH:
			if isinstance(data, str):
				return len(data)
			elif isinstance(data, int) or isinstance(data, float) or isinstance(data, Decimal):
				return len(str(data))
			else:
				raise DataKernelException(f'Cannot retrieve[key={name}, current={current_name}] from [{data}].')
		elif current_name == VariablePredefineFunctions.SUM:
			if isinstance(data, list):
				def to_decimal(value: Any) -> Decimal:
					if value is None:
						return Decimal(0)
					decimal_value = try_to_decimal(value)
					if decimal_value is None:
						raise DataKernelException(f'Cannot retrieve[key={name}, current={current_name}] from [{data}].')
					else:
						return decimal_value

				return ArrayHelper(data).reduce(lambda sum_value, value: sum_value + to_decimal(value), Decimal(0))
			else:
				raise DataKernelException(f'Cannot retrieve[key={name}, current={current_name}] from [{data}].')
		elif isinstance(data, dict):
			data = data.get(current_name)
		elif isinstance(data, list):
			# when parent is a list, value might be a primitive type or a list
			# for that value should be a list but is none, have to check the type by given function
			# this scenario is existing only on get value from raw trigger data
			# and get value from variables, and this value is read from raw trigger data
			value_is_list = is_list(names[:current_index + 1])

			def get_current_value(parent: Dict[str, Any]) -> Any:
				value = None if parent is None else parent.get(current_name)
				if value is None:
					if value_is_list:
						return []
					else:
						return value
				return value

			data = ArrayHelper(data).map(get_current_value).flatten().to_list()
		else:
			# cannot retrieve value from plain type variable
			raise DataKernelException(f'Cannot retrieve[key={name}, current={current_name}] from [{data}].')

		if data is None:
			# no need to go deeper
			if is_list(names[:current_index + 1]):
				return []
			else:
				return None
		elif isinstance(data, list) and len(data) == 0:
			# no need to go deeper
			return []

		# next loop
		current_index = current_index + 1
	# reaches target
	return data


# noinspection PyUnusedLocal
def always_none(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
	return None


def create_static_str(value: str) -> Callable[[PipelineVariables, PrincipalService], Any]:
	# noinspection PyUnusedLocal
	def get_static_str(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		return value

	return get_static_str


def create_snowflake_generator(prefix: str) -> Callable[[PipelineVariables, PrincipalService], Any]:
	# noinspection PyUnusedLocal
	def action(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		value = ask_snowflake_generator().next_id()
		return value if len(prefix) == 0 else f'{prefix}{value}'

	return action


def create_previous_trigger_data() -> Callable[[PipelineVariables, PrincipalService], Any]:
	# noinspection PyUnusedLocal
	def action(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		return variables.get_previous_trigger_data()

	return action


def create_from_previous_trigger_data(prefix, name: str) -> Callable[[PipelineVariables, PrincipalService], Any]:
	names = name.strip().split('.')

	# noinspection PyUnusedLocal
	def action(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		previous_data = variables.get_previous_trigger_data()
		value = get_value_from(name, names, lambda x: previous_data.get(x), variables.is_list_on_trigger)
		return value if len(prefix) == 0 else f'{prefix}{value}'

	return action


def create_get_value_from_variables(variables: PipelineVariables) -> Callable[[str], Any]:
	"""
	recursive is not supported
	"""

	def get_value(name: str) -> Any:
		if variables.has(name):
			return variables.find(name)
		else:
			return variables.find_from_current_data(name)

	return get_value


def create_is_list_from_variables(variables: PipelineVariables) -> Callable[[List[str]], bool]:
	def is_list(names: List[str]) -> bool:
		if variables.has(names[0]):
			return variables.is_list_on_variables(names)
		else:
			return variables.is_list_on_trigger(names)

	return is_list


def create_get_from_variables_with_prefix(prefix, name: str) -> Callable[[PipelineVariables, PrincipalService], Any]:
	names = name.strip().split('.')

	# noinspection PyUnusedLocal
	def action(variables: PipelineVariables, principal_service: PrincipalService) -> Any:
		value = get_value_from(
			name, names, create_get_value_from_variables(variables), create_is_list_from_variables(variables))
		return value if len(prefix) == 0 else f'{prefix}{value}'

	return action


def test_date(variable_name: str) -> Tuple[bool, Optional[date]]:
	if variable_name == VariablePredefineFunctions.NOW:
		return True, get_current_time_in_seconds()
	else:
		return is_date(variable_name, ask_all_date_formats())


# noinspection PyUnusedLocal
def get_date_from_variables(
		variables: PipelineVariables, principal_service: PrincipalService, variable_name: str
) -> Tuple[bool, Any, date]:
	value = get_value_from(
		variable_name, variable_name.strip().split('.'),
		create_get_value_from_variables(variables), create_is_list_from_variables(variables))
	if isinstance(value, date):
		return True, value, value
	parsed, parsed_date = is_date(value, ask_all_date_formats())
	return parsed, value, parsed_date


# noinspection PyUnusedLocal
def get_value_from_variables(
		variables: PipelineVariables, principal_service: PrincipalService, variable_name: str
) -> Tuple[bool, Any, Any]:
	value = get_value_from(
		variable_name, variable_name.strip().split('.'),
		create_get_value_from_variables(variables), create_is_list_from_variables(variables))
	return True, value, value


def compute_date_diff(
		function: VariablePredefineFunctions, end_date: date, start_date: date, variable_name: str
) -> int:
	if function == VariablePredefineFunctions.YEAR_DIFF:
		return year_diff(end_date, start_date)
	elif function == VariablePredefineFunctions.MONTH_DIFF:
		return month_diff(end_date, start_date)
	elif function == VariablePredefineFunctions.DAY_DIFF:
		return (truncate_time(end_date) - truncate_time(start_date)).days
	else:
		raise DataKernelException(f'Constant[{variable_name}] is not supported.')



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Dict, Union

from watchmen_data_kernel.common import DataKernelException
from watchmen_model.admin import FactorEncryptMethod


class Encryptor:
	@abstractmethod
	def accept(self, method: Union[FactorEncryptMethod, str]) -> bool:
		pass

	@abstractmethod
	def is_encrypted(self, value: Any) -> bool:
		"""
		value never be none or empty string
		"""
		pass

	@abstractmethod
	def do_encrypt(self, value: Any) -> Any:
		"""
		value never be none or empty string
		"""
		pass

	def encrypt(self, value: Any) -> Any:
		if value is None:
			return value
		elif isinstance(value, str) and len(value) == 0:
			return value
		elif self.is_encrypted(value):
			return value
		else:
			return self.do_encrypt(value)

	@abstractmethod
	def do_decrypt(self, value: Any) -> Any:
		"""
		value never be none or empty string
		"""
		pass

	def decrypt(self, value: Any) -> Any:
		if value is None:
			return value
		elif isinstance(value, str) and len(value) == 0:
			return value
		elif self.is_encrypted(value):
			return self.do_decrypt(value)
		else:
			return value

	# noinspection PyMethodMayBeStatic
	def should_ask_params(self) -> bool:
		return False

	@abstractmethod
	def get_key_type(self) -> str:
		pass

	def create_particular(self, params: Dict[str, Any]) -> Encryptor:
		raise DataKernelException('Method[create_particular] is not implemented on key type[{self.get_key_type()}].')



from __future__ import annotations

from datetime import date, datetime, time
from decimal import Decimal
from typing import Any

from watchmen_data_kernel.common import ask_all_date_formats, ask_full_datetime_formats, ask_time_formats, \
	DataKernelException
from watchmen_model.admin import Factor, FactorType
from watchmen_utilities import is_date, is_decimal, is_time


def cast_value_for_factor(value: Any, factor: Factor) -> Any:
	if value is None:
		return None

	factor_type = factor.type
	if factor_type in [
		FactorType.SEQUENCE, FactorType.NUMBER, FactorType.UNSIGNED, FactorType.FLOOR, FactorType.RESIDENTIAL_AREA,
		FactorType.AGE, FactorType.BIZ_SCALE
	]:
		parsed, decimal_value = is_decimal(value)
		if parsed:
			return decimal_value
		else:
			raise DataKernelException(
				f'Value[{value}] is incompatible with factor[name={factor.name}, type={factor_type}].')
	elif factor_type == FactorType.TEXT:
		if isinstance(value, str):
			return value
		elif isinstance(value, (int, float, Decimal, bool, date, time)):
			return str(value)
		else:
			raise DataKernelException(
				f'Value[{value}, type={type(value)}] is incompatible with '
				f'factor[name={factor.name}, type={factor_type}].')
	elif factor_type in [
		FactorType.ADDRESS, FactorType.ROAD, FactorType.COMMUNITY, FactorType.EMAIL, FactorType.PHONE,
		FactorType.MOBILE, FactorType.FAX, FactorType.OCCUPATION, FactorType.ID_NO
	]:
		if isinstance(value, str):
			return value
		elif isinstance(value, (int, float, Decimal)):
			return str(value)
		else:
			raise DataKernelException(
				f'Value[{value}, type={type(value)}] is incompatible with '
				f'factor[name={factor.name}, type={factor_type}].')
	# noinspection PyPep8
	elif factor_type in [
		FactorType.CONTINENT, FactorType.REGION, FactorType.COUNTRY, FactorType.PROVINCE, FactorType.CITY,
		FactorType.DISTRICT, FactorType.RESIDENCE_TYPE, FactorType.GENDER, FactorType.RELIGION, FactorType.NATIONALITY,
		FactorType.BIZ_TRADE, FactorType.ENUM
	]:
		if isinstance(value, str):
			return value
		elif isinstance(value, (int, Decimal)):
			return str(value)
		else:
			raise DataKernelException(
				f'Value[{value}, type={type(value)}] is incompatible with '
				f'factor[name={factor.name}, type={factor_type}].')
	elif factor_type == FactorType.FULL_DATETIME:
		# noinspection DuplicatedCode
		if isinstance(value, datetime):
			return value
		if isinstance(value, date):
			return datetime(year=value.year, month=value.month, day=value.day)
		if isinstance(value, time):
			raise DataKernelException(
				f'Value[{value}, type={type(value)}] is incompatible with '
				f'factor[name={factor.name}, type={factor_type}].')

		parsed, date_value = is_date(str(value), ask_full_datetime_formats())
		if parsed:
			return date_value
		else:
			raise DataKernelException(
				f'Value[{value}, type={type(value)}] is incompatible with '
				f'factor[name={factor.name}, type={factor_type}].')
	elif factor_type == FactorType.DATETIME:
		# noinspection DuplicatedCode
		if isinstance(value, datetime):
			return value
		if isinstance(value, date):
			return datetime(year=value.year, month=value.month, day=value.day)
		if isinstance(value, time):
			raise DataKernelException(
				f'Value[{value}, type={type(value)}] is incompatible with '
				f'factor[name={factor.name}, type={factor_type}].')

		parsed, date_value = is_date(str(value), ask_all_date_formats())
		if parsed:
			return date_value
		else:
			raise DataKernelException(
				f'Value[{value}, type={type(value)}] is incompatible with '
				f'factor[name={factor.name}, type={factor_type}].')
	elif factor_type in [
		FactorType.DATE, FactorType.DATE_OF_BIRTH
	]:
		if isinstance(value, datetime):
			return value.date()
		if isinstance(value, date):
			return value
		if isinstance(value, time):
			raise DataKernelException(
				f'Value[{value}, type={type(value)}] is incompatible with '
				f'factor[name={factor.name}, type={factor_type}].')

		parsed, date_value = is_date(value, ask_all_date_formats())
		if parsed:
			if isinstance(date_value, datetime):
				return date_value.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=None)
			else:
				return date_value
		else:
			raise DataKernelException(
				f'Value[{value}, type={type(value)}] is incompatible with '
				f'factor[name={factor.name}, type={factor_type}].')
	elif factor_type == FactorType.TIME:
		if isinstance(value, datetime):
			return value.time()
		if isinstance(value, date):
			raise DataKernelException(
				f'Value[{value}, type={type(value)}] is incompatible with '
				f'factor[name={factor.name}, type={factor_type}].')
		if isinstance(value, time):
			return value

		parsed, time_value = is_time(value, ask_time_formats())
		if parsed:
			return time_value
		else:
			raise DataKernelException(
				f'Value[{value}, type={type(value)}] is incompatible with '
				f'factor[name={factor.name}, type={factor_type}].')
	elif factor_type in [
		FactorType.YEAR, FactorType.HALF_YEAR, FactorType.QUARTER,
		FactorType.MONTH, FactorType.HALF_MONTH, FactorType.TEN_DAYS,
		FactorType.WEEK_OF_YEAR, FactorType.WEEK_OF_MONTH, FactorType.HALF_WEEK,
		FactorType.DAY_OF_MONTH, FactorType.DAY_OF_WEEK, FactorType.DAY_KIND,
		FactorType.HOUR, FactorType.HOUR_KIND,
		FactorType.MINUTE, FactorType.SECOND, FactorType.MILLISECOND,
		FactorType.AM_PM
	]:
		# TODO strictly validation is needed or not?
		parsed, decimal_value = is_decimal(value)
		if parsed:
			return decimal_value
		else:
			raise DataKernelException(
				f'Value[{value}] is incompatible with factor[name={factor.name}, type={factor_type}].')
	elif factor_type == FactorType.BOOLEAN:
		if isinstance(value, bool):
			return value
		elif isinstance(value, (int, float, Decimal)):
			return value != 0
		elif isinstance(value, str):
			v = value.strip().lower()
			if v == 't' or v == 'y' or v == 'yes' or v == 'true':
				return True
			elif v == 'f' or v == 'n' or v == 'no' or v == 'false':
				return False
		raise DataKernelException(
			f'Value[{value}, type={type(value)}] is incompatible with '
			f'factor[name={factor.name}, type={factor_type}].')
	else:
		raise DataKernelException(f'Factor type[{factor_type}] is not supported.')



from typing import List, Optional

from watchmen_model.common import TenantId
from watchmen_model.system import Tenant
from .cache_manager import get_tenant_by_id_cache
from .internal_cache import InternalCache


# noinspection DuplicatedCode
class TenantCache:
	"""
	tenant cache will not impact other caches
	"""

	def __init__(self):
		self.byIdCache = InternalCache(cache=get_tenant_by_id_cache)

	def put(self, tenant: Tenant) -> Optional[Tenant]:
		return self.byIdCache.put(tenant.tenantId, tenant)

	def get(self, tenant_id: TenantId) -> Optional[Tenant]:
		return self.byIdCache.get(tenant_id)

	def remove(self, tenant_id: TenantId) -> Optional[Tenant]:
		return self.byIdCache.remove(tenant_id)

	def all(self) -> List[Tenant]:
		return list(self.byIdCache.values())

	def clear(self) -> None:
		self.byIdCache.clear()


tenant_cache = TenantCache()



from re import findall
from typing import Any, List, Tuple, Union

from math import ceil, floor

from watchmen_data_kernel.common import DataKernelException
from watchmen_model.admin import FactorEncryptMethod
from watchmen_utilities import ArrayHelper
from .encryptor import Encryptor


class CenterMasker(Encryptor):
	def __init__(self, digits: int):
		self.digits = digits
		if digits == 3:
			self.method = FactorEncryptMethod.MASK_CENTER_3
		elif digits == 5:
			self.method = FactorEncryptMethod.MASK_CENTER_5
		else:
			raise DataKernelException(f'Only 3 or 5 digits center mask is supported, current is [{digits}].')

	def accept(self, method: Union[FactorEncryptMethod, str]) -> bool:
		return method == self.method

	def is_encrypted(self, value: Any) -> bool:
		"""
		always returns false
		"""
		return False

	def do_encrypt_as_normal(self, value: str, length: int) -> str:
		digits = self.digits
		remains = length - digits
		if remains % 2 == 1:
			pad_count = floor(remains / 2)
			return f'{value[:pad_count + 1]}{"*" * digits}{value[length - pad_count:]}'
		else:
			pad_count = int(remains / 2)
			return f'{value[:pad_count]}{"*" * digits}{value[length - pad_count:]}'

	# noinspection PyMethodMayBeStatic
	def replace_from_left(self, segment: str, remain_chars_count: int) -> Tuple[int, str]:
		char_count = len(segment)
		if char_count >= remain_chars_count:
			return 0, f'{"*" * remain_chars_count}{segment[remain_chars_count:]}'
		else:
			return remain_chars_count - char_count, '*' * char_count

	# noinspection PyMethodMayBeStatic
	def replace_from_right(self, segment: str, remain_chars_count: int) -> Tuple[int, str]:
		char_count = len(segment)
		if char_count >= remain_chars_count:
			return 0, f'{segment[:0 - remain_chars_count]}{"*" * remain_chars_count}'
		else:
			return remain_chars_count - char_count, '*' * char_count

	def do_encrypt(self, value: Any) -> Any:
		digits = self.digits

		value = str(value)
		length = len(value)
		if length <= digits:
			# mask all
			return '*' * length

		# last part is empty, ignored
		segments = ArrayHelper(findall(r'((\d+|\D*)?)', value)[:-1]).map(lambda x: x[0]).to_list()
		decimal_char_count = ArrayHelper(segments).filter(lambda x: x.isdecimal()) \
			.reduce(lambda count, x: count + len(x), 0)
		if decimal_char_count < digits:
			# no enough decimal characters, mask as normal string
			return self.do_encrypt_as_normal(value, length)

		decimal_count = len([segment for segment in segments if segment.isdecimal()])
		if decimal_count == 1:
			# only one part
			replaced_segments = []
			for segment in segments:
				if segment.isdecimal():
					replaced_segments.append(f'{"*" * digits}{segment[digits:]}')
				else:
					replaced_segments.append(segment)
			return ''.join(replaced_segments)

		# at least 2 decimal parts
		central_index = ceil((decimal_count + 1) / 2)
		index = 0

		replaced_segments: List[str] = []
		current_index = central_index
		remain_count = digits

		for segment in segments:
			if segment.isdecimal():
				index = index + 1
				if index == current_index:
					remain_count, replaced = self.replace_from_left(segment, remain_count)
					replaced_segments.append(replaced)
					if remain_count != 0:
						# try to replace next decimal segment
						current_index = current_index + 1
				else:
					replaced_segments.append(segment)
			else:
				replaced_segments.append(segment)

		if remain_count == 0:
			return ''.join(replaced_segments)
		else:
			# still not enough characters, try to look backward
			current_index = central_index - 1
			while current_index > 0:
				segments = replaced_segments
				replaced_segments = []
				index = 0
				for segment in segments:
					if segment.isdecimal():
						index = index + 1
						if index == current_index:
							remain_count, replaced = self.replace_from_right(segment, remain_count)
							replaced_segments.append(replaced)
							if remain_count != 0:
								# look backward again
								current_index = current_index - 1
							else:
								# break
								current_index = 0
						else:
							replaced_segments.append(segment)
					else:
						replaced_segments.append(segment)
			return ''.join(replaced_segments)

	def do_decrypt(self, value: str) -> str:
		# center mask cannot be decrypted
		return value

	def get_key_type(self) -> str:
		return self.method.value



from base64 import b64decode, b64encode
from typing import Any, Dict, Union

from watchmen_data_kernel.common import DataKernelException
from watchmen_model.admin import FactorEncryptMethod
from watchmen_utilities import is_blank, is_empty
from .encryptor import Encryptor


class AESEncryptor(Encryptor):
	def __init__(self, key: str, iv: str):
		self.key = key.encode('utf-8')
		self.iv = iv.encode('utf-8')

	def accept(self, method: Union[FactorEncryptMethod, str]) -> bool:
		return method == FactorEncryptMethod.AES256_PKCS5_PADDING

	def is_encrypted(self, value: Any) -> bool:
		if value is None:
			return False
		elif isinstance(value, str) and value.startswith('{AES}'):
			return True
		else:
			return False

	# noinspection PyMethodMayBeStatic
	def ask_aes(self):
		# lazy load
		# noinspection PyPackageRequirements
		from Crypto.Cipher import AES
		return AES.new(self.key, AES.MODE_CFB, self.iv)

	def do_encrypt(self, value: Any) -> Any:
		if value is None or (isinstance(value, str) and is_empty(value)):
			return value
		value = str(value)
		return '{AES}' + b64encode(self.ask_aes().encrypt(value.encode('utf-8'))).decode()

	def do_decrypt(self, value: str) -> str:
		if is_empty(value):
			return value
		# remove prefix and decrypt
		return self.ask_aes().decrypt(b64decode(value[5:])).decode('utf-8')

	def should_ask_params(self) -> bool:
		return True

	def get_key_type(self) -> str:
		return FactorEncryptMethod.AES256_PKCS5_PADDING.value

	def create_particular(self, params: Dict[str, Any]) -> Encryptor:
		key = params.get('key')
		if is_blank(key) or len(key) != 32:
			raise DataKernelException(f'Parameter key[{key}] should be 32 digits.')
		iv = params.get('iv')
		if is_blank(iv) or len(iv) != 16:
			raise DataKernelException(f'Parameter iv[{iv}] should be 16 digits.')
		return AESEncryptor(key, iv)



from __future__ import annotations

from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple

from watchmen_model.admin import Factor, FactorType, Topic
from watchmen_utilities import ArrayHelper, is_blank


class PipelineVariables:
	def __init__(
			self, previous_data: Optional[Dict[str, Any]], current_data: Optional[Dict[str, Any]],
			topic: Optional[Topic]
	):
		self.previousData = previous_data
		self.currentData = current_data
		self.variables: Dict[str, Any] = {}
		# only variables from trigger data will record its factor name here
		# key is variable key, value is factor name
		self.variables_from: Dict[str, str] = {}
		self.topic = topic

	def is_list_on_trigger(self, names: List[str]) -> bool:
		if self.topic is None:
			# no topic declared, false
			return False

		name = ArrayHelper(names).join('.')
		factor: Optional[Factor] = ArrayHelper(self.topic.factors).find(lambda x: x.name == name)
		if factor is None:
			return False
		else:
			return factor.type == FactorType.ARRAY

	def is_list_on_variables(self, names: List[str]) -> bool:
		factor_name = self.variables_from.get(names[0])
		if is_blank(factor_name):
			return False
		else:
			factor_name = factor_name + '.' + ArrayHelper(names[1:]).join('.')
			factor: Optional[Factor] = ArrayHelper(self.topic.factors).find(lambda x: x.name == factor_name)
			if factor is None:
				return False
			else:
				return factor.type == FactorType.ARRAY

	def trace_variable(self, name: str) -> Tuple[bool, Optional[str]]:
		if name in self.variables_from:
			return True, self.variables_from.get(name)
		else:
			return False, None

	def put_with_from(self, name: str, value: Any, from_: str):
		self.put(name, value)
		self.variables_from[name] = from_

	def put(self, name: str, value: Any):
		self.variables[name] = value
		if name in self.variables_from:
			del self.variables_from[name]

	def has(self, name: str) -> bool:
		return name in self.variables

	def find(self, name: str) -> Optional[Any]:
		return self.variables.get(name)

	def find_from_current_data(self, name: str) -> Optional[Any]:
		return self.currentData.get(name)

	def has_previous_trigger_data(self) -> bool:
		return self.previousData is not None

	def get_previous_trigger_data(self) -> Optional[Dict[str, Any]]:
		return self.previousData

	def clone(self) -> PipelineVariables:
		cloned = PipelineVariables(self.previousData, self.currentData, self.topic)
		cloned.variables_from = deepcopy(self.variables_from)
		cloned.variables = deepcopy(self.variables)
		return cloned

	def clone_all(self) -> PipelineVariables:
		cloned = PipelineVariables(deepcopy(self.previousData), deepcopy(self.currentData), self.topic)
		cloned.variables_from = deepcopy(self.variables_from)
		cloned.variables = deepcopy(self.variables)
		return cloned

	def backward_to_previous(self):
		"""
		not cloned, assume it will not be changed
		"""
		backed = PipelineVariables(None, self.previousData, self.topic)
		backed.variables_from = deepcopy(self.variables_from)
		backed.variables = self.variables
		return backed



from typing import Any, Dict, List

from watchmen_model.admin import Factor, Topic
from watchmen_utilities import ArrayHelper, is_blank


class FlattenFactor:
	def __init__(self, factor: Factor):
		self.factor = factor
		self.factorName = '' if is_blank(factor.name) else factor.name.strip()
		self.names = self.factorName.split('.')

	def get_factor(self):
		return self.factor

	def get_names(self):
		return self.names

	def flatten(self, root: Dict[str, Any]) -> Any:
		data = root
		value = None
		for name in self.names:
			value = data.get(name)
			if value is None:
				# break and set to root
				root[self.factorName] = None
				return None
			else:
				data = value
		# set to root
		root[self.factorName] = value
		return value


def parse_flatten_factors(topic: Topic) -> List[FlattenFactor]:
	return ArrayHelper(topic.factors).filter(lambda x: x.flatten).map(lambda x: FlattenFactor(x)).to_list()



from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from watchmen_auth import PrincipalService
from watchmen_data_kernel.encryption import Encryptor, find_encryptor, register_encryptor
from watchmen_model.admin import Factor, FactorEncryptMethod, Topic
from watchmen_utilities import ArrayHelper, is_blank, is_not_blank

"""
design of encryption.
looping by encrypt methods, leads unnecessary visits of raw topic hierarchy. for this reason, encrypt methods are 
parsed to groups.
for example, here are several properties need to be encrypted:
1. a,
2. b.c,
3. b.d,
4. b.e.f.g.h.i
5. b.e.f.g.h.j
looping by methods, leads unnecessary visits on "b" and "b.e.f.g.h" (one time for each in this case). 
if there are array and many items, may cause performance issues.
now methods will be parsed to:
1. group "a"
	1.1 factor "a"
2. group "b"
	2.1 group "c"
		2.1.1 factor "c"
	2.2 group "d"
		2.2.1 factor "d"
	2.3 group "e"
		2.3.1 group "f"
			2.3.1.1 group "g"
				2.3.1.1.1 group "h"
					2.3.1.1.1.1 factor "i"
					2.3.1.1.1.2 factor "j"
now, loop by groups, there is no duplicated visit anymore.
"""


def ask_encryptor(method: Union[FactorEncryptMethod, str], principal_service: PrincipalService) -> Encryptor:
	encryptor = find_encryptor(method)
	if encryptor.should_ask_params():
		key_type = encryptor.get_key_type()
		tenant_id = principal_service.get_tenant_id()
		particular_key = f'{tenant_id}-{key_type}'
		particular_encryptor = find_encryptor(particular_key)
		if particular_encryptor is not None:
			return particular_encryptor
		# to avoid loop dependency
		from watchmen_data_kernel.meta import KeyStoreService
		key_store = KeyStoreService(principal_service).find_by_type(key_type, tenant_id)
		if key_store is None:
			# no special key store declared, use global one
			return encryptor
		else:
			particular_encryptor = encryptor.create_particular(key_store.params)
			register_encryptor(particular_key, particular_encryptor)
			return particular_encryptor
	else:
		return encryptor


class EncryptFactor:
	def __init__(self, factor: Factor):
		self.factor = factor
		self.factorName = '' if is_blank(factor.name) else factor.name.strip()
		self.names = self.factorName.split('.')

	def pop_first_name(self):
		self.names = self.names[1:]

	def get_encrypt_method(self):
		return self.factor.encrypt


class EncryptFactorGroup:
	factors: Optional[List[EncryptFactor]] = None
	groups: Optional[List[EncryptFactorGroup]] = None

	def __init__(self, name: str, factors: List[EncryptFactor]):
		self.name = name
		# in reality, zero or one factor.
		# if there is one, name is same as group's, and will not contain group anymore
		self.factors = ArrayHelper(factors).filter(lambda x: len(x.names) == 1).to_list()
		groups = ArrayHelper(factors).filter(lambda x: len(x.names) > 1) \
			.each(lambda x: x.pop_first_name()) \
			.group_by(lambda x: x.names[0])
		self.groups = ArrayHelper(list(groups.items())) \
			.map(lambda x: EncryptFactorGroup(name=x[0], factors=x[1])).to_list()

	def encrypt(self, data: Dict[str, Any], principal_service: PrincipalService) -> None:
		value = data.get(self.name)
		if value is None:
			return
		if isinstance(value, dict):
			ArrayHelper(self.groups).each(lambda x: x.encrypt(value, principal_service))
		elif isinstance(value, list):
			def each(item):
				ArrayHelper(self.groups).each(lambda x: x.encrypt(item, principal_service))

			ArrayHelper(value).each(lambda x: each(x))
		else:
			data[self.name] = encrypt(value, self.factors[0].get_encrypt_method(), principal_service)

	def decrypt(self, data: Dict[str, Any], principal_service: PrincipalService) -> None:
		value = data.get(self.name)
		if value is None:
			return
		if isinstance(value, dict):
			ArrayHelper(self.groups).each(lambda x: x.decrypt(value, principal_service))
		elif isinstance(value, list):
			def each(item):
				ArrayHelper(self.groups).each(lambda x: x.decrypt(item, principal_service))

			ArrayHelper(value).each(lambda x: each(x))
		else:
			data[self.name] = decrypt(value, self.factors[0].get_encrypt_method(), principal_service)


def parse_encrypt_factors(topic: Topic) -> List[EncryptFactorGroup]:
	groups = ArrayHelper(topic.factors) \
		.filter(lambda x: x.encrypt is not None and x.encrypt != FactorEncryptMethod.NONE) \
		.map(lambda x: EncryptFactor(x)) \
		.filter(lambda x: is_not_blank(x.factorName)) \
		.group_by(lambda x: x.names[0])

	return ArrayHelper(list(groups.items())) \
		.map(lambda x: EncryptFactorGroup(name=x[0], factors=x[1])).to_list()


def encrypt(value: Any, method: FactorEncryptMethod, principal_service: PrincipalService) -> Any:
	# do encryption
	# check it is encrypted or not first
	# data read from topic, and write again, might be encrypted already
	if method == FactorEncryptMethod.NONE:
		return value
	else:
		return ask_encryptor(method, principal_service).encrypt(value)


def decrypt(value: Any, method: FactorEncryptMethod, principal_service: PrincipalService) -> Any:
	# do encryption
	# check it is encrypted or not first
	# data read from topic, and write again, might be encrypted already
	if method == FactorEncryptMethod.NONE:
		return value
	else:
		return ask_encryptor(method, principal_service).decrypt(value)



from .data_service import TopicDataEntityHelper, TopicDataService, TopicStructureService, TopicTrigger
from .raw_data_service import RawTopicDataEntityHelper, RawTopicDataService
from .regular_data_service import RegularTopicDataEntityHelper, RegularTopicDataService
from .topic_storage import build_topic_data_storage



from abc import abstractmethod
from typing import Any, Dict

from watchmen_data_kernel.topic_schema import TopicSchema
from watchmen_model.admin import TopicKind
from watchmen_model.pipeline_kernel import TopicDataColumnNames
from watchmen_storage import EntityRow, EntityShaper
from .factor_column_mapper import TopicFactorColumnMapper


class TopicShaper(EntityShaper):
	def __init__(self, schema: TopicSchema):
		self.schema = schema
		self.mapper = self.create_factor_column_mapper(schema)

	def get_schema(self) -> TopicSchema:
		return self.schema

	def is_synonym(self) -> bool:
		return self.get_schema().get_topic().kind == TopicKind.SYNONYM

	@abstractmethod
	def create_factor_column_mapper(self, schema: TopicSchema) -> TopicFactorColumnMapper:
		pass

	def get_mapper(self) -> TopicFactorColumnMapper:
		return self.mapper

	# noinspection PyMethodMayBeStatic
	def serialize_fix_columns(self, data: Dict[str, Any]) -> EntityRow:
		return {
			TopicDataColumnNames.ID.value: data.get(TopicDataColumnNames.ID.value),
			TopicDataColumnNames.TENANT_ID.value: data.get(TopicDataColumnNames.TENANT_ID.value),
			TopicDataColumnNames.INSERT_TIME.value: data.get(TopicDataColumnNames.INSERT_TIME.value),
			TopicDataColumnNames.UPDATE_TIME.value: data.get(TopicDataColumnNames.UPDATE_TIME.value)
		}

	# noinspection PyMethodMayBeStatic
	def deserialize_fix_columns(self, row: EntityRow) -> Dict[str, Any]:
		return {
			TopicDataColumnNames.ID.value: row.get(TopicDataColumnNames.ID.value),
			TopicDataColumnNames.TENANT_ID.value: row.get(TopicDataColumnNames.TENANT_ID.value),
			TopicDataColumnNames.INSERT_TIME.value: row.get(TopicDataColumnNames.INSERT_TIME.value),
			TopicDataColumnNames.UPDATE_TIME.value: row.get(TopicDataColumnNames.UPDATE_TIME.value)
		}

	def serialize_factor(self, data: Dict[str, Any], factor_name: str, row: EntityRow) -> None:
		value = data.get(factor_name)
		column_name = self.get_mapper().get_column_name(factor_name)
		row[column_name] = value

	def deserialize_column(self, row: EntityRow, column_name: str, data: Dict[str, Any]) -> None:
		value = row.get(column_name)
		factor_name = self.get_mapper().get_factor_name(column_name)
		data[factor_name] = value



from typing import Callable

from watchmen_data_kernel.common import ask_storage_echo_enabled, DataKernelException
from watchmen_model.system import DataSource, DataSourceType
from watchmen_storage import TopicDataStorageSPI


def build_mysql_storage(data_source: DataSource) -> Callable[[], TopicDataStorageSPI]:
	from watchmen_storage_mysql import StorageMySQLConfiguration, MySQLDataSourceParams
	configuration = StorageMySQLConfiguration(data_source, MySQLDataSourceParams(echo=ask_storage_echo_enabled()))
	return lambda: configuration.create_topic_data_storage()


def build_oracle_storage(data_source: DataSource) -> Callable[[], TopicDataStorageSPI]:
	from watchmen_storage_oracle import StorageOracleConfiguration, OracleDataSourceParams
	configuration = StorageOracleConfiguration(data_source, OracleDataSourceParams(echo=ask_storage_echo_enabled()))
	return lambda: configuration.create_topic_data_storage()


def build_mongodb_storage(data_source: DataSource) -> Callable[[], TopicDataStorageSPI]:
	from watchmen_storage_mongodb import StorageMongoConfiguration, MongoDataSourceParams
	configuration = StorageMongoConfiguration(data_source, MongoDataSourceParams(echo=ask_storage_echo_enabled()))
	return lambda: configuration.create_topic_data_storage()


def build_mssql_storage(data_source: DataSource) -> Callable[[], TopicDataStorageSPI]:
	from watchmen_storage_mssql import StorageMSSQLConfiguration, MSSQLDataSourceParams
	configuration = StorageMSSQLConfiguration(data_source, MSSQLDataSourceParams(echo=ask_storage_echo_enabled()))
	return lambda: configuration.create_topic_data_storage()


def build_postgresql_storage(data_source: DataSource) -> Callable[[], TopicDataStorageSPI]:
	from watchmen_storage_postgresql import StoragePostgreSQLConfiguration, PostgreSQLDataSourceParams
	configuration = StoragePostgreSQLConfiguration(
		data_source, PostgreSQLDataSourceParams(echo=ask_storage_echo_enabled()))
	return lambda: configuration.create_topic_data_storage()


def build_oss_storage(data_source: DataSource) -> Callable[[], TopicDataStorageSPI]:
	from watchmen_storage_oss import StorageOssConfiguration, OssDataSourceParams
	configuration = StorageOssConfiguration(
		data_source, OssDataSourceParams(echo=ask_storage_echo_enabled()))
	return lambda: configuration.create_topic_data_storage()


def build_s3_storage(data_source: DataSource) -> Callable[[], TopicDataStorageSPI]:
	from watchmen_storage_s3 import StorageS3Configuration, S3DataSourceParams
	configuration = StorageS3Configuration(
		data_source, S3DataSourceParams(echo=ask_storage_echo_enabled()))
	return lambda: configuration.create_topic_data_storage()


def build_topic_data_storage(data_source: DataSource) -> Callable[[], TopicDataStorageSPI]:
	if data_source.dataSourceType == DataSourceType.MYSQL:
		return build_mysql_storage(data_source)
	if data_source.dataSourceType == DataSourceType.ORACLE:
		return build_oracle_storage(data_source)
	if data_source.dataSourceType == DataSourceType.MONGODB:
		return build_mongodb_storage(data_source)
	if data_source.dataSourceType == DataSourceType.MSSQL:
		return build_mssql_storage(data_source)
	if data_source.dataSourceType == DataSourceType.POSTGRESQL:
		return build_postgresql_storage(data_source)
	if data_source.dataSourceType == DataSourceType.OSS:
		return build_oss_storage(data_source)
	if data_source.dataSourceType == DataSourceType.S3:
		return build_s3_storage(data_source)

	raise DataKernelException(
		f'Topic data storage[id={data_source.dataSourceId}, name={data_source.name} type={data_source.dataSourceType}] '
		f'is not supported yet.')



from typing import Any, Callable

from watchmen_model.admin import PipelineAction

ActionDefinedAs = Callable[[], Any]


def parse_action_defined_as(action: PipelineAction) -> ActionDefinedAs:
	defined_as = action.dict()
	return lambda: defined_as



from typing import List, Optional, Tuple

from watchmen_model.common import TenantId
from watchmen_model.system import KeyStore
from .cache_manager import get_key_store_by_type_cache
from .internal_cache import InternalCache


class KeyStoreCache:
	"""
	key store cache will not impact other caches
	"""

	def __init__(self):
		self.byTypeCache = InternalCache(cache=get_key_store_by_type_cache)

	def put(self, key_store: KeyStore) -> Optional[KeyStore]:
		return self.byTypeCache.put(f'{key_store.tenantId}-{key_store.keyType}', key_store)

	def declare_not_existing(self, key_type: str, tenant_id: TenantId) -> Optional[KeyStore]:
		key = f'{tenant_id}-{key_type}'
		return self.byTypeCache.put(key, key)

	def get(self, key_type: str, tenant_id: TenantId) -> Tuple[bool, Optional[KeyStore]]:
		"""
		first identify if it is hit, second is key store if exists.
		returns (True, None) when it is declared not existing
		"""
		key = f'{tenant_id}-{key_type}'
		value = self.byTypeCache.get(key)
		if value is None:
			return False, None
		elif value == key:
			return True, None
		else:
			return True, value

	def remove(self, key_type: str, tenant_id: TenantId) -> Optional[KeyStore]:
		data_source = self.byTypeCache.remove(f'{tenant_id}-{key_type}')
		return data_source

	def all(self) -> List[KeyStore]:
		return list(self.byTypeCache.values())

	def clear(self) -> None:
		self.byTypeCache.clear()


key_store_cache = KeyStoreCache()



from typing import Optional

from watchmen_auth import PrincipalService
from watchmen_data_kernel.cache import CacheService
from watchmen_data_kernel.common import DataKernelException
from watchmen_data_kernel.external_writer import find_external_writer_create, register_external_writer_creator
from watchmen_meta.common import ask_meta_storage, ask_snowflake_generator
from watchmen_meta.system import ExternalWriterService as ExternalWriterStorageService
from watchmen_model.common import ExternalWriterId
from watchmen_model.system import ExternalWriter


def register_external_writer(external_writer: ExternalWriter) -> None:
	create = find_external_writer_create(external_writer.type)
	if create is None:
		raise DataKernelException(f'Creator not found for external writer[{external_writer.dict()}].')
	register_external_writer_creator(external_writer.writerCode, create())


class ExternalWriterService:
	def __init__(self, principal_service: PrincipalService):
		self.principalService = principal_service

	def find_by_id(self, writer_id: ExternalWriterId) -> Optional[ExternalWriter]:
		external_writer = CacheService.external_writer().get(writer_id)
		if external_writer is not None:
			if external_writer.tenantId != self.principalService.get_tenant_id():
				raise DataKernelException(
					f'External writer[id={writer_id}] not belongs to '
					f'current tenant[id={self.principalService.get_tenant_id()}].')
			register_external_writer(external_writer)
			return external_writer

		storage_service = ExternalWriterStorageService(
			ask_meta_storage(), ask_snowflake_generator(), self.principalService)
		storage_service.begin_transaction()
		try:
			# noinspection PyTypeChecker
			external_writer: ExternalWriter = storage_service.find_by_id(writer_id)
			if external_writer is None:
				return None

			CacheService.external_writer().put(external_writer)
			register_external_writer(external_writer)
			return external_writer
		finally:
			storage_service.close_transaction()



from typing import Callable, Dict, Optional

from watchmen_model.system import ExternalWriterType
from .external_writer_registry import BuildExternalWriter

CreateExternalWriterBuilder = Callable[[], BuildExternalWriter]


class ExternalWriterBuilderRegistry:
	creators: Dict[ExternalWriterType, CreateExternalWriterBuilder] = {}

	def register(self, writer_type: ExternalWriterType, creator_create: CreateExternalWriterBuilder) -> None:
		self.creators[writer_type] = creator_create

	def find(self, writer_type: ExternalWriterType) -> Optional[CreateExternalWriterBuilder]:
		return self.creators.get(writer_type)


external_writer_registry = ExternalWriterBuilderRegistry()


def register_external_writer_create(
		writer_type: ExternalWriterType, creator_create: CreateExternalWriterBuilder) -> None:
	external_writer_registry.register(writer_type, creator_create)


def find_external_writer_create(writer_type: ExternalWriterType) -> Optional[CreateExternalWriterBuilder]:
	return external_writer_registry.find(writer_type)



from typing import Any, Union

from watchmen_model.admin import FactorEncryptMethod
from .encryptor import Encryptor


class MailMasker(Encryptor):
	def accept(self, method: Union[FactorEncryptMethod, str]) -> bool:
		return method == FactorEncryptMethod.MASK_MAIL

	def is_encrypted(self, value: Any) -> bool:
		if value is None:
			return False
		elif isinstance(value, str) and value.startswith('*****@'):
			return True
		else:
			return False

	def do_encrypt(self, value: Any) -> Any:
		value = str(value)
		pos = value.find('@')
		if pos > 0:
			return f'*****{value[pos:]}'
		else:
			return value

	def do_decrypt(self, value: str) -> str:
		# mail mask cannot be decrypted
		return value

	def get_key_type(self) -> str:
		return FactorEncryptMethod.MASK_MAIL.value



from .ask_action_defined_as import parse_action_defined_as
from .ask_from_memory import parse_parameter_in_memory, parse_prerequisite_defined_as, parse_prerequisite_in_memory, \
	ParsedMemoryConstantParameter, ParsedMemoryParameter, PrerequisiteDefinedAs, PrerequisiteTest
from .ask_from_storage import parse_condition_for_storage, parse_mapping_for_storage, parse_parameter_for_storage, \
	ParsedStorageCondition, ParsedStorageMapping, PossibleParameterType
from .time_utils import now, spent_ms
from .topic_utils import ask_topic_data_entity_helper
from .utils import always_none, get_value_from
from .variables import PipelineVariables



from .exception import DataKernelException
from .settings import ask_all_date_formats, ask_cache_enabled, ask_cache_heart_beat_enabled, \
	ask_cache_heart_beat_interval, ask_date_formats, ask_datetime_formats, ask_encrypt_aes_params, \
	ask_full_datetime_formats, ask_ignore_default_on_raw, ask_replace_topic_to_storage, ask_storage_echo_enabled, \
	ask_sync_topic_to_storage, ask_time_formats, ask_topic_snapshot_scheduler_heart_beat_interval, ask_trino_enabled



from logging import getLogger
from typing import List, Tuple

from pydantic import BaseSettings

from watchmen_utilities import ArrayHelper

logger = getLogger(__name__)


class KernelSettings(BaseSettings):
	STORAGE_ECHO: bool = False

	FULL_DATETIME_FORMATS: List[str] = [
		'%Y%m%d%H%M%S%f', '%d%m%Y%H%M%S%f', '%m%d%Y%H%M%S%f',  # 14 or more digits,
	]
	DATETIME_FORMATS: List[str] = [
		'%Y%m%d%H%M%S', '%d%m%Y%H%M%S', '%m%d%Y%H%M%S',  # 14 digits,
		'%Y%m%d%H%M', '%d%m%Y%H%M', '%m%d%Y%H%M'  # 12 digits
	]  # all digits, other characters are prohibitive
	DATE_FORMATS: List[str] = [
		'%Y%m%d', '%d%m%Y', '%m%d%Y',  # 8 digits
	]  # all digits, other characters are prohibitive
	TIME_FORMATS: List[str] = [
		'%H%M%S',  # 6 digits
		'%H%M'  # 4 digits
	]  # all digits, other characters are prohibitive
	ABANDON_DATE_TIME_ON_PARSE_FAIL: bool = False

	ENCRYPT_AES_KEY: str = 'hWmZq4t7w9z$C&F)J@NcRfUjXn2r5u8x'  # AES key of factor encryption
	ENCRYPT_AES_IV: str = 'J@NcRfUjXn2r5u8x'  # AES iv of factor encryption

	IGNORE_DEFAULT_ON_RAW: bool = True  # default value settings will be ignored on the raw topic

	KERNEL_CACHE: bool = True  # enable kernel cache, keep it enabled in production
	KERNEL_CACHE_HEART_BEAT: bool = True  # enable kernel cache heart beat
	KERNEL_CACHE_HEART_BEAT_INTERVAL: int = 60  # kernel cache heart beat interval, in seconds

	TOPIC_SNAPSHOT_SCHEDULER_HEART_BEAT_INTERVAL: int = 30  # topic snapshot scheduler heart beat interval, in seconds

	SYNC_TOPIC_TO_STORAGE: bool = False  # sync topic change to storage entity
	REPLACE_TOPIC_TO_STORAGE: bool = False  # force replace existing topic entity (drop and recreate)
	TRINO: bool = True  # trino

	class Config:
		# secrets_dir = '/var/run'
		env_file = '.env'
		env_file_encoding = 'utf-8'
		case_sensitive = True


settings = KernelSettings()
logger.info(f'Data kernel settings[{settings.dict()}].')

full_datetime_formats = list(settings.FULL_DATETIME_FORMATS)
datetime_formats = list(settings.DATETIME_FORMATS)
date_formats = list(settings.DATE_FORMATS)
all_date_formats = ArrayHelper(list(settings.FULL_DATETIME_FORMATS)) \
	.grab(*settings.DATETIME_FORMATS) \
	.grab(*settings.DATE_FORMATS) \
	.to_list()
time_formats = list(settings.TIME_FORMATS)


def ask_storage_echo_enabled() -> bool:
	return settings.STORAGE_ECHO


def ask_full_datetime_formats() -> List[str]:
	return full_datetime_formats


def ask_datetime_formats() -> List[str]:
	return datetime_formats


def ask_date_formats() -> List[str]:
	return date_formats


def ask_all_date_formats() -> List[str]:
	return all_date_formats


def ask_time_formats() -> List[str]:
	return time_formats


def ask_abandon_date_time_on_parse_fail() -> bool:
	return settings.ABANDON_DATE_TIME_ON_PARSE_FAIL


def ask_encrypt_aes_params() -> Tuple[str, str]:
	"""
	key, iv
	"""
	return settings.ENCRYPT_AES_KEY, settings.ENCRYPT_AES_IV


# noinspection DuplicatedCode
def ask_ignore_default_on_raw() -> bool:
	return settings.IGNORE_DEFAULT_ON_RAW


def ask_cache_enabled() -> bool:
	return settings.KERNEL_CACHE


def ask_cache_heart_beat_enabled() -> bool:
	return settings.KERNEL_CACHE_HEART_BEAT


def ask_cache_heart_beat_interval() -> int:
	return settings.KERNEL_CACHE_HEART_BEAT_INTERVAL


def ask_topic_snapshot_scheduler_heart_beat_interval() -> int:
	return settings.TOPIC_SNAPSHOT_SCHEDULER_HEART_BEAT_INTERVAL


def ask_sync_topic_to_storage() -> bool:
	return settings.SYNC_TOPIC_TO_STORAGE


def ask_replace_topic_to_storage() -> bool:
	return settings.REPLACE_TOPIC_TO_STORAGE


def ask_trino_enabled() -> bool:
	return settings.TRINO



from typing import Any, Dict, List, Optional

from watchmen_data_kernel.topic_schema import TopicSchema
from watchmen_model.admin import Factor, FactorType, TopicKind
from watchmen_model.pipeline_kernel import TopicDataColumnNames
from watchmen_storage import EntityCriteriaExpression, EntityRow, EntityShaper
from watchmen_utilities import ArrayHelper
from .data_entity_helper import TopicDataEntityHelper
from .data_service import TopicDataService
from .factor_column_mapper import TopicFactorColumnMapper
from .shaper import TopicShaper


class RawTopicFactorColumnMapper(TopicFactorColumnMapper):
	def get_factors(self, schema: TopicSchema) -> List[Factor]:
		if schema.get_topic().kind == TopicKind.SYNONYM:
			# top level factors are required
			return ArrayHelper(schema.get_topic().factors).filter(lambda x: '.' not in x.name).to_list()
		else:
			# only flatten factors are required
			return ArrayHelper(schema.get_flatten_factors()).map(lambda x: x.get_factor()).to_list()


class RawTopicShaper(TopicShaper):
	def create_factor_column_mapper(self, schema: TopicSchema) -> TopicFactorColumnMapper:
		return RawTopicFactorColumnMapper(schema)

	def serialize(self, data: Dict[str, Any]) -> EntityRow:
		if self.is_synonym():
			row: EntityRow = {}
			ArrayHelper(self.get_mapper().get_factor_names()).each(lambda x: self.serialize_factor(data, x, row))
		else:
			row = self.serialize_fix_columns(data)
			row[TopicDataColumnNames.RAW_TOPIC_DATA.value] = data.get(TopicDataColumnNames.RAW_TOPIC_DATA.value)
			ArrayHelper(self.get_mapper().get_factor_names()).each(lambda x: self.serialize_factor(data, x, row))
		return row

	def deserialize(self, row: EntityRow) -> Dict[str, Any]:
		if self.is_synonym():
			temp: Dict[str, Any] = {}
			# copy columns to temp dict
			ArrayHelper(self.get_mapper().get_column_names()).each(lambda x: self.deserialize_column(row, x, temp))
			data: Dict[str, Any] = {}
			# copy factors which on top level except for object and array type, just like they are flatten
			for k, v in temp.items():
				factor: Optional[Factor] = ArrayHelper(self.get_schema().get_topic().factors) \
					.find(lambda x: x.name == k)
				if factor is not None and factor.type != FactorType.OBJECT and factor.type != FactorType.ARRAY:
					data[k] = v
			# copy whole object as data column
			data[TopicDataColumnNames.RAW_TOPIC_DATA.value] = temp
		else:
			data = self.deserialize_fix_columns(row)
			data[TopicDataColumnNames.RAW_TOPIC_DATA.value] = row.get(TopicDataColumnNames.RAW_TOPIC_DATA.value)
			ArrayHelper(self.get_mapper().get_column_names()).each(lambda x: self.deserialize_column(row, x, data))
		return data


class RawTopicDataEntityHelper(TopicDataEntityHelper):
	def create_entity_shaper(self, schema: TopicSchema) -> EntityShaper:
		return RawTopicShaper(schema)

	def is_versioned(self) -> bool:
		return False

	def find_version(self, data: Dict[str, Any]) -> int:
		"""
		always return -1
		"""
		return -1

	def build_version_criteria(self, data: Dict[str, Any]) -> Optional[EntityCriteriaExpression]:
		return None

	def assign_version(self, data: Dict[str, Any], version: int) -> None:
		"""
		do nothing, raw topic has no version column
		"""
		pass


class RawTopicDataService(TopicDataService):
	def try_to_wrap_to_topic_data(self, data: Dict[str, Any]) -> Dict[str, Any]:
		self.delete_reversed_columns(data)

		def wrap_flatten_factor(factor: Factor, from_data: Dict[str, Any], to_data: Dict[str, Any]) -> None:
			name = factor.name
			# copy flatten value to wrapped data
			to_data[name] = from_data.get(name)
			if name.find('.') != -1 and name in from_data:
				del from_data[name]

		wrapped_data = {TopicDataColumnNames.RAW_TOPIC_DATA.value: data}
		# retrieve flatten factors
		flatten_factors = self.schema.get_flatten_factors()
		ArrayHelper(flatten_factors) \
			.map(lambda x: x.get_factor()) \
			.each(lambda x: wrap_flatten_factor(x, data, wrapped_data))
		return wrapped_data

	def try_to_unwrap_from_topic_data(self, topic_data: Dict[str, Any]) -> Dict[str, Any]:
		unwrapped_data = {}

		# remove flatten factors
		reserved_keys = [
			TopicDataColumnNames.ID.value,
			TopicDataColumnNames.TENANT_ID.value,
			TopicDataColumnNames.INSERT_TIME.value,
			TopicDataColumnNames.UPDATE_TIME.value
		]
		for key, value in topic_data.items():
			if key in reserved_keys:
				unwrapped_data[key] = value

		if TopicDataColumnNames.RAW_TOPIC_DATA.value in topic_data:
			pure_data = topic_data.get(TopicDataColumnNames.RAW_TOPIC_DATA.value)
			if pure_data is not None:
				for key, value in pure_data.items():
					unwrapped_data[key] = value
		return unwrapped_data



from typing import List, Optional

from watchmen_auth import PrincipalService
from watchmen_data_kernel.cache import CacheService
from watchmen_meta.common import ask_meta_storage, ask_snowflake_generator
from watchmen_meta.system import TenantService as TenantStorageService
from watchmen_model.common import TenantId
from watchmen_model.system import Tenant


class TenantService:
	def __init__(self, principal_service: PrincipalService):
		self.principalService = principal_service

	def find_by_id(self, tenant_id: TenantId) -> Optional[Tenant]:
		tenant = CacheService.tenant().get(tenant_id)
		if tenant is not None:
			return tenant

		storage_service = TenantStorageService(ask_meta_storage(), ask_snowflake_generator(), self.principalService)
		storage_service.begin_transaction()
		try:
			# noinspection PyTypeChecker
			tenant: Tenant = storage_service.find_by_id(tenant_id)
			if tenant is None:
				return None

			CacheService.tenant().put(tenant)
			return tenant
		finally:
			storage_service.close_transaction()

	def find_all(self) -> List[Tenant]:
		storage_service = TenantStorageService(ask_meta_storage(), ask_snowflake_generator(), self.principalService)
		storage_service.begin_transaction()
		try:
			# noinspection PyTypeChecker
			return storage_service.find_all()
		finally:
			storage_service.close_transaction()
